program Show_PCX_File;
{****************************************************************************
** Demonstration of loading a PCX file and displaying it on screen         **
**  by Steven H Don                                                        **
**                                                                         **
** This support type 5, 256 colour PCX files with a maximum resolution     **
** of 320x200.                                                             **
**                                                                         **
** For questions, feel free to e-mail me.                                  **
**                                                                         **
**    shd@earthling.net                                                    **
**    http://shd.cjb.net                                                   **
**                                                                         **
****************************************************************************}

uses Dos, Crt;

type PCXHeader = record
  Manufacturer : Byte;
  Version      : Byte;
  Encoding     : Byte;
  BitsPerPixel : Byte;
  xMin         : Integer;
  yMin         : Integer;
  xMax         : Integer;
  yMax         : Integer;
  Other        : Array [1..116] of Char;
end;

type BufferType = Array [0..63999] of Byte;

var
  {An offscreen buffer}
  Buffer        : ^BufferType;
  FileName      : String;
  Palette       : Array [0..255, 0..2] of Byte;
  x, y          : Integer;

procedure VgaScreen; assembler;
{This sets the display to VGA 320x200 in 256 colours}
asm
  mov ax, 13h
  int 10h
end;

{This sets a DAC register to a specific Red Green Blue-value}
procedure SetDAC(DAC, R, G, B : Byte);
begin
  Port[$3C8] := DAC;
  Port[$3C9] := R;
  Port[$3C9] := G;
  Port[$3C9] := B;
end;

procedure TextScreen; assembler;
{This resets the display to text mode}
asm
  mov ax, 3h
  int 10h
end;

{Sets a pixel in the buffer to a specific colour}
procedure PutPixel (x, y : Word; c : Byte);
begin
  if (x < 320) and (y < 200) then
    Buffer^ [y shl 8 + y shl 6 + x] := c;
end;

{Assigns memory for a buffer}
function MakeBuffer : Boolean;
begin
  {Check if enough memory is available}
  if MaxAvail < 64000 then begin
    {If not, return false}
    MakeBuffer := false;
    exit;
  end;
  New (Buffer);
end;

{Clears the buffer}
procedure ClearBuffer;
begin
  fillchar (Buffer^, 64000, 0);
end;

{Displays the buffer}
procedure DisplayBuffer;
begin
  move (Buffer^,          {from buffer}
        Mem [$A000:0000], {to screen}
        64000);           {64000 pixels}
end;

procedure LoadPCX(f : String);
{
  This loads in PCX File and stores it in the buffer.
}
var
  PCXLen                         : LongInt;
  DOffSet, OffSet, Width, Height : Word;
  HowMany, DataByte, Counter     : Byte;
  Header                         : PCXHeader;
  PCXFileOpened                  : Boolean;
  Palette                        : Array [0..255,0..2] of Byte;
  PCXBuffer                      : Array [1..512] of Byte;
  nextPointer                    : Word;
  PCXFile                        : File;
  x, y                           : Integer;

  {This function reads the next byte from the PCX file}
  function NextByte : Byte;
  begin
    {Check to see whether next block should be read}
    if nextPointer < 1 then begin
      {If so, read it and reset the pointer}
      BlockRead (PCXFile, PCXBuffer [1], SizeOf (PCXBuffer));
      nextPointer := 1;
    end;

    {Return byte in buffer}
    NextByte := PCXBuffer [nextPointer];

    {Adjust pointer}
    inc (nextPointer);
    if nextPointer > 512 then NextPointer := 0;
  end;

begin
  nextPointer := 0;
  FileMode := 0; {Read only}
  {Check to see whether the file exists and can be opened}
  {$I-}
     PCXFileOpened := true;
     Assign (PCXFile, f);
     Reset (PCXFile, 1);
     if IOResult <> 0 then begin
       f := f + '.PCX';
       Assign (PCXFile, f);
       Reset (PCXFile, 1);
     end;
     Reset (PCXFile, 1);
     if IOResult <> 0 then PCXFileOpened := false;
  {$I+}
  if not PCXFileOpened then exit;

  {Read in header information}
  BlockRead (PCXFile, Header, SizeOf (Header));
  {Check to see whether we can display it}
  if Header.Version <> 5 then begin
    {If other version than 5, don't display}
    Close (PCXFile);
    exit;
  end;
  if (Header.xMax > 319) or (Header.xMin < 0)
  or (Header.yMax > 199) or (Header.yMin < 0) then begin
    {If it doesn't fit on-screen, don't display}
    Close(PCXFile);
    exit;
  end;

  {Load in the palette}
  Seek (PCXFile, FileSize (PCXFile) - 769);
  {Read in identifier}
  BlockRead (PCXFile, DataByte, 1);
  if DataByte <> 12 then begin
    {If there is no palette, don't display}
    Close (PCXFile);
    exit;
  end;

  BlockRead (PCXFile, Palette[0,0], 768);
  {Divide all palette entries by 4}
  for Counter := 0 to 255 do begin
    Palette[Counter,0] := Palette[Counter, 0] SHR 2;
    Palette[Counter,1] := Palette[Counter, 1] SHR 2;
    Palette[Counter,2] := Palette[Counter, 2] SHR 2;
    SetDAC(Counter,
           Palette [Counter, 0],
           Palette [Counter, 1],
           Palette [Counter, 2]);
  end;

  {Go back to start of graphic data}
  Seek (PCXFile, 128);

  x := Header.xMin;
  y := Header.yMin;
  inc (Header.xMax);
  inc (Header.yMax);

  {Decode and display graphics}
  While (y < Header.yMax) do begin
    {Read next byte}
    DataByte := NextByte;

    {Reset counter}
    HowMany := 1;

    {If it is encoded, extract the count information and read in the colour byte}
    if (DataByte and $C0) = $C0 THEN begin
      HowMany := DataByte and $3F;
      DataByte := NextByte;
    end;

    {Display it}
    for Counter := 1 to HowMany do begin
      {Store pixel in the buffer}
      PutPixel (x, y, DataByte);
      {Next pixel}
      inc (x);

      {If End of Line reached, next line}
      if x = Header.xMax then begin
        inc (y);
        x := Header.xMin;
      end;

    end;
  end;
  Close (PCXFile);
end;

begin

  {Check if a filename was passed as a parameter, otherwise ask for one}
  if ParamCount > 0 then
    filename := ParamStr (1)
  else begin
    write ('Enter filename:'); readln (filename);
  end;

  VgaScreen;

  {Assign memory for the buffer}
  MakeBuffer;

  {Load the PCX File}
  LoadPCX(filename);

  {Display the buffer}
  DisplayBuffer;

  Readln;
  TextScreen;

end.